;;; A patch work with code from:
;;;
;;; http://sourceforge.net/mailarchive/message.php?msg_id=6301700
;;; http://www.scsh.net/resources/sunet.html
;;; http://www.scsh.net/cgi-bin/wiki.cgi?SendXmlSuspend

;;; Emilio Lopes <eclig@gmx.net>, 2006-03-17.

(define attribute-rule
  `(@ 
    ((*default*                         ; local override for attributes
      . ,(lambda (attr-key . value)
           (if (null? value)
               (list #\space attr-key "='" attr-key "'")
               (list #\space attr-key "='" value #\')))))
    . ,(lambda (trigger . value) (cons '@  value))))

(define (format-element tag attribs contents)
  (list "<" tag attribs
        (if (pair? contents)
            (list  ">" contents "</" tag ">")
            "/>")
        #\newline))

;; Predicate for attributes in sxml, from SUNet.
(define (sxml-attribute? thing)
  (and (pair? thing)
       (eq? '@ (car thing))))

;; Returns the attribute list after the initial '@. For our
;; convenience, it ignores values that are not sxml-attributes.
;; Also from SUNet.
(define (sxml-attribute-attributes thing)
  (if (sxml-attribute? thing)
      (cdr thing)
      '()))

(define default-rule
  `(*default*
    . ,(lambda (tag . elems)
         (if (and (pair? elems)
                  (sxml-attribute? (car elems)))
             (format-element tag (sxml-attribute-attributes (car elems)) (cdr elems))
             (format-element tag '() elems)))))

(define text-rule
  `(*text*
    . ,(lambda (trigger str) 
	 (if (string? str) (string->goodHTML str) str))))

(define processing-instruction-rule
  `(*PI* *preorder*
         . ,(lambda (tag . elems)
              `(,(string-append "<?" (symbol->string (car elems)) " ")
                ,@(cdr elems)
                "?>"
                #\newline))))

(define top-rule
  `(*TOP*
    . ,(lambda (tag . elems)
         (list elems))))

(define xml-rules
  `(,attribute-rule
    ,default-rule
    ,text-rule
    ,processing-instruction-rule
    ,top-rule))

(define (sxml->xml sxml-expr)
  (SRV:send-reply
   (pre-post-order sxml-expr xml-rules)))
