;;; Simple example of class precedence in Tiny-CLOS
;; Emilio Lopes <eclig@gmx.net>

,open srfi-1
,open srfi-13

(define <named-class>
  (make-class (list <class>)
              (list 'name)))

(define (named-class? obj)
  (eq? <named-class> (class-of obj)))

(add-method initialize
  (make-method (list <named-class>)
    (lambda (cnm obj initargs)
      (initialize-slots obj initargs)
      (cnm))))

(define initialize-slots
  (lambda (object initargs)
    (let ((not-there (list 'shes-not-there)))
      (for-each (lambda (slot)
                  (let ((name (car slot)))
                    (let ((value  (getl initargs name not-there)))
                      (if (eq? value not-there)
                          'do-nothing
                          (slot-set! object name value)))))
                (class-slots (class-of object))))))

(define getl
  (lambda (initargs name . not-found)
    (letrec ((scan (lambda (tail)
                     (cond ((null? tail)
                            (if (pair? not-found)
                                (car not-found)
                                (error "GETL couldn't find" name)))
                           ((eq? (car tail) name)
                            (cadr tail))
                           (else
                            (scan (cddr tail)))))))
      (scan initargs))))

(define class-name (make-generic))

(add-method class-name
  (make-method (list <named-class>)
    (lambda (cnm named-class)
      (let ((name (slot-ref named-class 'name)))
        (if (not (null? name))
            name
            "**anonymous class**")))))



(define (print-class-cpl class)
  (display (string-join (map class-name
                             (filter named-class? (class-cpl class)))
                        " -> "))
  (newline))


(define <O>
  (make <named-class> 'direct-supers (list <object>) 'name "Class O"))

(define <F>
  (make <named-class> 'direct-supers (list <O>) 'name "Class F"))

(define <E>
  (make <named-class> 'direct-supers (list <O>) 'name "Class E"))

(define <D>
  (make <named-class> 'direct-supers (list <O>) 'name "Class D"))

(define <C>
  (make <named-class> 'direct-supers (list <D> <F>) 'name "Class C"))

(define <B>
  (make <named-class> 'direct-supers (list <D> <E>) 'name "Class B"))

(define <A>
  (make <named-class> 'direct-supers (list <B> <C>) 'name "Class A"))

(print-class-cpl <A>)
;; With the standard Tiny-CLOS linearization, the above will print:
;;
;;   Class A -> Class B -> Class C -> Class D -> Class F -> Class E -> Class O
;;
;; With the C3 linearization one gets:
;;
;;   Class A -> Class B -> Class C -> Class D -> Class E -> Class F -> Class O
